# Generated by Django 2.2.7 on 2020-05-18 20:06

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Page',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('url', models.CharField(max_length=256, unique=True)),
                ('public', models.BooleanField(default=True)),
                ('template', models.CharField(default='en', max_length=256)),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='created_pages', to=settings.AUTH_USER_MODEL)),
                ('updated_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('upper', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='nodes', to='paipage.Page')),
            ],
            options={
                'unique_together': {('upper', 'url')},
            },
        ),
        migrations.CreateModel(
            name='PageText',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('language', models.CharField(max_length=16)),
                ('title', models.CharField(max_length=256)),
                ('keywords', models.CharField(max_length=256)),
                ('description', models.TextField()),
                ('text_short', models.TextField()),
                ('text_full', models.TextField()),
                ('created_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='created_texts', to=settings.AUTH_USER_MODEL)),
                ('page', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='texts', to='paipage.Page')),
                ('updated_by', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import cachetools

from django.http import HttpResponse

from paipage.const import HTML_EXT
from paipage import config

__all__ = [
    'TemplateHandler',
    'should_show', 'render_page_short', 'render_page_full',
]


class TemplateHandler():
    # PAF elements list with additional page settings
    form_elements = None

    def __init__(self, template, page, text, params):
        # Argument list may be changed, don't override or use *args, **kwargs
        self.template = template
        self.page = page
        self.text = text
        self.params = params
        self.lang = params['lang']
        self.request = params['request']

    @staticmethod
    @cachetools.cached(cache=cachetools.TTLCache(maxsize=16, ttl=10))
    def get_children_for(page):
        return list(page.children.all())

    def get_children(self):
        return self.get_children_for(self.page)

    def should_show(self):
        return len(self.text.text_full) > 1

    def get_short(self):
        # Must return HTML string
        return self.text.text_short

    def render_full(self):
        # Must return django.http.HttpResponse
        return HttpResponse(self.params.render_file(self.template))

    def __repr__(self):
        return f'{self.__class__.__name__} for "{self.page.url}" {self.lang}'


def _ensure_pg(page):
    template = page.template
    if not template:
        template = config.template_page_default
    return template


def _ensure_lo(page):
    layout = page.layout
    if not layout:
        layout = config.template_layout_default
    return layout


def get_handler(params, page, text):
    template = _ensure_pg(page)
    if template in config.template_handlers:
        handler_class = config.template_handlers[template]
    else:
        handler_class = TemplateHandler

    return handler_class(
        template, page, text, params
    )


def should_show(params, page, text):
    handler = get_handler(params, page, text)
    return handler.should_show()


def render_page_short(params, page, text):
    handler = get_handler(params, page, text)
    return handler.get_short()


def render_page_full(params, page, text):
    handler = get_handler(params, page, text)
    template = handler.template
    layout = _ensure_lo(page)
    layout += HTML_EXT

    params.update({
        'current': page,
        'handler': handler,
        'text': text,
        'upper': page.upper,
        'children': handler.get_children(),
        'layout_template': layout,
    })
    return handler.render_full()



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import sys
import operator

from django.conf import settings


def is_migration():
    return 'makemigrations' in sys.argv or 'migrate' in sys.argv


class SettingObj:
    _serialise = []
    _repr = []
    _init_done = False

    def __init__(self):
        self._init_done = True

    def __setattr__(self, name, value):
        if self._init_done:
            if name not in self.__dict__:
                raise ValueError(f'Trying to set attr "{name}" for {self}')
        self.__dict__[name] = value

    def as_dict(self):
        return {key: getattr(self, key) for key in self._serialise}

    def __repr__(self):
        fileds = ', '.join(
            list(map(lambda key: f'{key}={self.__dict__[key]!r}', self._repr))
        )
        return f'{self.__class__.__name__}({fileds})'


class PluginInfo(SettingObj):
    _serialise = (
        'title', 'description', 'version', 'license', 'author', 'contact',
        'name', 'path', 'template_page_list', 'template_layout_list',
    )
    _repr = ['name', 'version']

    def __init__(self, plugin_name, plugin_dir):
        self.title = '_none_'
        self.description = '_none_'
        self.version = '_unknown_'
        self.license = '_unknown_'
        self.author = '_unknown_'
        self.contact = '_unknown_'
        self.name = plugin_name
        self.path = plugin_dir
        self.template_page_list = []
        self.template_layout_list = []
        self.template_handlers = {}
        self._init_done = True


class PluginMeta(SettingObj):
    _serialise = (
        'name', 'enabled',
    )
    _repr = ['name', 'enabled']

    def __init__(self, name, enabled=False):
        self.name = name
        self.enabled = enabled
        self._init_done = True


def dedupe(items, key=None, attr=None):
    if isinstance(key, str):
        key_getter = lambda x: x[key]
    elif key is None:
        if attr is not None:
            key_getter = operator.attrgetter(attr)
        else:
            key_getter = lambda x: x
    else:
        key_getter = key

    seen = set()
    for item in items:
        _key = key_getter(item)
        if _key not in seen:
            yield item
            seen.add(_key)


RE_PG_FILE = r'^(pg-.+)\.html$'
RE_LO_FILE = r'^(lo-.+)\.html$'
RE_PG_NAME = r'^(pg-.+)$'
RE_LO_NAME = r'^(lo-.+)$'

HTML_EXT = '.html'
PATH_INDEX = 'index'

LANG_NO = '_'

LANG_KEY = 'lang'

REQUIRED_STRINGS = (
    'name',  # lang name
    'languages',  # lang selection menu
    'error404',  # for bad URLs
    'errorNoLang',  # for pages with no text in current lang
    'errorNoLangOther',  # "check other languages"
    'back',  # "Back" link
)


class Language:
    def __init__(self, code, strings):
        self.code = code
        for key in REQUIRED_STRINGS:
            assert key in strings, f'No string "{key}" for lang "{code}"'
        self._strings = strings

    def as_dict(self):
        return self._strings

    def __setitem__(self, key, item):
        self._strings[key] = item

    def __getitem__(self, key):
        if key in self._strings:
            return self._strings[key]
        else:
            txt = f'Missing string "{key}"'
            if settings.DEBUG:
                raise KeyError(txt)
            else:
                return f'[ {txt} ]'


lang_en = Language(
    code='en',
    strings=dict(
        name='English',
        languages='Other languages',
        error404='Page not found',
        error500='Server error :(',
        errorNoLang='Not available in this language',
        errorNoLangOther='You can read the page in another language:',
        back='Back',
    ),
)
lang_ru = Language(
    code='ru',
    strings=dict(
        name='Русский',
        languages='Другие языки',
        error404='Страница не найдена',
        error500='Ошибка сервера :(',
        errorNoLang='Недоступно на этом языке',
        errorNoLangOther='Вы можете посмотреть на другом языке:',
        back='Назад',
    ),
)



__title__ = 'PaiPaige Simple plugin'
__version__ = '1.0'
__license__ = 'GPL-3.0'
__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'
description = 'Provides simple yet handy templates.'

import json

import cachetools

from django.shortcuts import get_object_or_404

from paipage import TemplateHandler
from paipage.models import Page


class PgLabeled(TemplateHandler):
	form_elements = [
		{
			'type': 'str',
			'max_len': '32',
			'code': 'label',
			'name': 'Label',
		}, {
			'type': 'str',
			'max_len': '128',
			'code': 'tags',
			'name': 'Tags',
			# TODO: add PAF type "words" to remove bad symbols
		},
	]

	def get_short(self):
		result = self.text.text_short
		label = self.page.other_settings_.get('label', '')
		if len(label) > 0:
			result += f'<br>{label}'
		return result


class PgNested(TemplateHandler):
	@staticmethod
	@cachetools.cached(cache=cachetools.TTLCache(maxsize=16, ttl=10))
	def get_children_for(page):
		cluster_settings = json.loads(page.other_settings)
		target = cluster_settings.get('target', None)
		if target is None:
			target = page
		else:
			target = get_object_or_404(Page, pk=target)

		# TODO: Filter by lang
		result = list(target.children.all())

		# Filter by tags if set
		embrace = lambda x: f' {x} '
		tags = cluster_settings.get('tags', '')
		tags = [embrace(tag) for tag in tags.split(' ') if len(tag) > 0]
		for tag in tags:
			result = [
				pg for pg in result
				if tag in embrace(pg.other_settings_.get('tags', ''))
			]

		# Order by label
		result.sort(
			reverse=cluster_settings.get('desc', True),
			key=lambda pg: pg.other_settings_.get('label', '')
		)

		# Limit by count
		count = cluster_settings.get('count', 0)
		if count > 0:
			result = result[:count]

		return result

	def should_show(self):
		return len(self.get_children()) > 0


class PgCluster(PgNested):
	form_elements = [
		{
			'type': 'select',
			'subtype': 'page',
			'allow_clear': True,
			'code': 'target',
			'name': 'Target page cluster',
		}, {
			'type': 'str',
			'max_len': '128',
			'code': 'tags',
			'name': 'Tags filter',
			# TODO: add PAF type "words" to remove bad symbols
		}, {
			'type': 'int',
			'min': '0',
			'code': 'count',
			'name': 'Max count',
			'desc': '0 value means no limit',
		}, {
			'type': 'bool',
			'code': 'desc',
			'name': 'Label order',
			'on': 'descending',
			'off': 'ascending',
			'default': True,
		},
	]


class PgContents(PgNested):

	def get_short(self):
		self.params.update({
			'children': list(self.page.children.all()),
		})
		text = ''
		if len(self.text.text_short) > 1:
			text = self.text.text_short

		template = f'''
{{% import "_utils.html" as utils with context %}}
{{{{ utils.page_contents(children, 2) }}}}
{text}
'''
		return self.params.render_str(template)


template_handlers = {
	'pg-usual': PgLabeled,
	'pg-cluster': PgCluster,
	'pg-contents': PgContents,
}



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

from jinja2 import Environment, FileSystemLoader


def safe_json(input):
    return input.replace('</script>', '<\\/script>')


def page_by_url(path):
    from paipage.models import Page
    page = Page.objects.filter(url=path).first()
    if page is None:
        raise ValueError(f'Missing page_by_url("{path}")')
    return page


def make_env(template_dirs, config):
    env = Environment(loader=FileSystemLoader(template_dirs))
    env.filters['safe_json'] = safe_json
    env.globals.update({
        'config': config,
        'page_by_url': page_by_url,
    })
    return env



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

from django.apps import AppConfig

from .config import configure_final


class PaiPaiConfig(AppConfig):
    name = 'paipage'

    def ready(self):
        configure_final()



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import json

from django.views import View
from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.contrib.admin.views.decorators import staff_member_required
from django.conf import settings

from paipage import config

from .params import Params


class AdminkaMainView(View):
    @method_decorator(staff_member_required)
    def get(self, request):
        params = Params(request)
        params['site_name'] = config.site_name
        params['debug'] = settings.DEBUG
        return HttpResponse(params.render_file('am-main'))



# Generated by Django 2.2.7 on 2020-06-11 17:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('paipage', '0002_auto_20200519_1746'),
    ]

    operations = [
        migrations.CreateModel(
            name='GlobalSetting',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=64, unique=True)),
                ('value', models.TextField(blank=True)),
                ('updated', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.AlterField(
            model_name='page',
            name='layout',
            field=models.CharField(blank=True, default='', max_length=256),
        ),
        migrations.AlterField(
            model_name='page',
            name='template',
            field=models.CharField(blank=True, default='', max_length=256),
        ),
    ]



__title__ = 'PaiPaige Primary plugin'
__version__ = '1.0'
__license__ = 'GPL-3.0'
__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'
description = 'Defines Adminka, commonly used macros, and other basic things.'



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import os
import sys
import importlib
import shutil
import re

from . import const
from .const import PluginInfo, PluginMeta
from .jinja2 import make_env

__all__ = [
    'config',
    'configure',
    'configure_final',
]


class Config(const.SettingObj):
    _serialise = [
        'site_name',
        # Default values that may be overriden by DB settings
        'language_default', 'language_available',
        'template_page_default', 'template_layout_default',
        'coded_plugin_enabled', 'plugins_meta',
        # Settings
        'plugin_paths', 'logger',
        # Generated values
        'plugins_installed',
        'template_page_list', 'template_layout_list', 'template_handlers', 'jinja2',
        'common_css', 'template_dirs', 'static_dirs',
    ]
    _repr = ['site_name']

    def __init__(self):
        for name in self._serialise:
            self.__dict__[name] = None
        self._init_done = True

    def get_plugin_meta_plain(self):
        return [
            plugin_meta.as_dict()
            for plugin_meta in self.plugins_meta
        ]


config = Config()


def configure(
    site_name,
    language_default,
    language_available,
    template_page_default='pg-pure',
    template_layout_default='lo-pure',
    plugin_paths=[],
    plugin_enabled=['simple'],
    logger=None,
):

    '''
        Basic setup
    '''
    assert isinstance(site_name, str) and len(site_name) > 0
    config.site_name = site_name

    config.template_page_default = template_page_default
    config.template_layout_default = template_layout_default
    config.template_page_list = []
    config.template_layout_list = []

    for x in language_available:
        assert isinstance(x, const.Language)
    config.language_available = {x.code: x for x in language_available}
    assert isinstance(language_default, str)
    assert language_default in config.language_available
    config.language_default = language_default

    config.plugin_paths = plugin_paths
    config.coded_plugin_enabled = plugin_enabled

    if logger is None:
        import logging
        logger = logging.getLogger('paipage')
        logger.setLevel(logging.DEBUG)
        console_handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter('%(asctime)s %(levelname)s: %(message)s')
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    config.logger = logger


def attr2attr(src, dst, src_name, dst_name=None):
    if dst_name is None:
        dst_name = src_name
    if hasattr(src, src_name):
        setattr(dst, dst_name, getattr(src, src_name))


def load_plugin_code(plugin_info):
    plugin_name = plugin_info.name
    plugin_dir = plugin_info.path
    code_file = os.path.join(plugin_dir, '__init__.py')
    if os.path.exists(code_file):
        module_name = f'pai_plugin_{plugin_name}'
        module = importlib.machinery.SourceFileLoader(
            module_name, code_file).load_module()
        attr2attr(module, plugin_info, 'template_handlers')
        attr2attr(module, plugin_info, '__title__', 'title')
        attr2attr(module, plugin_info, '__version__', 'version')
        attr2attr(module, plugin_info, '__license__', 'license')
        attr2attr(module, plugin_info, '__author__', 'author')
        attr2attr(module, plugin_info, '__contact__', 'contact')
        attr2attr(module, plugin_info, 'description', 'description')


def collect_plugins(search_dir):
    search_dir = os.path.join(search_dir, 'plugins')
    if not os.path.exists(search_dir):
        return
    for plugin_name in os.listdir(search_dir):
        plugin_dir = os.path.join(search_dir, plugin_name)
        if not os.path.isdir(plugin_dir):
            continue

        plugin_info = PluginInfo(plugin_name, plugin_dir)

        # Load list of templates
        location = os.path.join(plugin_dir, 'templates')
        if os.path.exists(location):
            for filename in os.listdir(location):
                match = re.findall(const.RE_PG_FILE, filename)
                if len(match) == 1:
                    plugin_info.template_page_list.append(match[0])
                match = re.findall(const.RE_LO_FILE, filename)
                if len(match) == 1:
                    plugin_info.template_layout_list.append(match[0])

        load_plugin_code(plugin_info)

        config.plugins_installed[plugin_name] = plugin_info


def enable_plugins():
    config.common_css = []
    config.template_dirs = []
    config.static_dirs = []
    config.template_handlers = {}

    def add_common_css(plugin_dir, plugin_name, filename):
        src = os.path.join(plugin_dir, 'static', filename)
        res = f'_pai-{plugin_name}-{filename}'
        if os.path.exists(src):
            # TODO: copy to STATIC_ROOT if not DEBUG
            shutil.copyfile(src, os.path.join(plugin_dir, 'static', res))
            config.common_css.append(res)

    for plugins_meta in config.plugins_meta:
        if not plugins_meta.enabled:
            continue
        plugin_name = plugins_meta.name
        if plugin_name not in config.plugins_installed:
            raise ValueError(f'Missing enabled plugin "{plugin_name}"')
            continue
        plugin_info = config.plugins_installed[plugin_name]
        plugin_dir = plugin_info.path

        location = os.path.join(plugin_dir, 'templates')
        if os.path.exists(location):
            config.template_dirs.append(location)

        location = os.path.join(plugin_dir, 'static')
        if os.path.exists(location):
            config.static_dirs.append(location)

        add_common_css(plugin_dir, plugin_name, 'main.css')

        for name in plugin_info.template_page_list:
            if name not in config.template_page_list:
                config.template_page_list.append(name)

        for name in plugin_info.template_layout_list:
            if name not in config.template_layout_list:
                config.template_layout_list.append(name)

        config.template_handlers.update(plugin_info.template_handlers)


def configure_final():
    if const.is_migration():
        return

    from django.conf import settings
    from .models import GlobalSetting

    # # Plugins loading
    config.plugins_installed = {}
    for app in settings.INSTALLED_APPS:
        app_path = os.path.dirname(sys.modules[app].__file__)
        collect_plugins(app_path)
    for path in config.plugin_paths:
        collect_plugins(path)

    # # Plugins enabling
    saved_plugins = GlobalSetting.get_json('plugins_meta', [])
    if not saved_plugins:
        saved_plugins = [
            {'name': name, 'enabled': True}
            for name in config.coded_plugin_enabled
        ]

    if saved_plugins:
        saved_plugins = [
            PluginMeta(**kwargs) for kwargs in saved_plugins
        ]
    else:
        saved_plugins = [
            PluginMeta(name, True)
            for name in config.coded_plugin_enabled
        ]

    config.plugins_meta = []
    config.plugins_meta.append(PluginMeta('primary', True))
    config.plugins_meta.extend(saved_plugins)
    config.plugins_meta.extend(
        PluginMeta(name, False)
        for name in config.plugins_installed
    )
    config.plugins_meta = list(const.dedupe(config.plugins_meta, attr='name'))

    enable_plugins()

    config.jinja2 = make_env(template_dirs=config.template_dirs, config=config)

    # # Django settings
    settings.STATICFILES_DIRS.extend(config.static_dirs)
    settings.SESSION_ENGINE = 'django.contrib.sessions.backends.signed_cookies'



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

from django.contrib import admin
from django.urls import path, include

from . import views

urlpatterns = [
	path('adminka/', views.AdminkaMainView.as_view()),
	path('adminka/plugins', views.AdminkaPluginsView.as_view()),
	path('adminka/structure', views.AdminkaStructureView.as_view()),
	path('adminka/page/<int:pk>', views.AdminkaPageView.as_view()),
	path('adminka/page/create', views.AdminkaPageView.as_view()),
	path('adminka/preview/<int:ind>', views.PagePreView.as_view()),
	path('adminka/text/<int:pk>/<slug:lang>', views.AdminkaPageTextView.as_view()),

	path('', views.PageView.as_view()),
	path('<slug:path>', views.PageView.as_view()),
	path('_lang/<slug:lang>', views.ChangeLangView.as_view()),
	path('<slug:lang>/<slug:path>', views.PageView.as_view()),
]



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import json

from django.db.models import Count
from django.views import View
from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.contrib.admin.views.decorators import staff_member_required

from paipage import config
from paipage.config import configure_final
from paipage import const
from paipage.models import Page, GlobalSetting

from .params import Params


class AdminkaPluginsView(View):
    @method_decorator(staff_member_required)
    def get(self, request):
        params = Params(request)
        params.scripted['pluginsInstalled'] = {
            name: config.plugins_installed[name].as_dict()
            for name in config.plugins_installed
        }
        params.scripted['pluginsMeta'] = config.get_plugin_meta_plain()

        params.scripted['template_layout_default'] = config.template_layout_default
        params.scripted['template_page_default'] = config.template_page_default

        params.scripted['page_templates'] = {
            d['template']: d['count']
            for d in Page.objects.values('template').annotate(count=Count('template'))
        }
        params.scripted['layout_templates'] = {
            d['layout']: d['count']
            for d in Page.objects.values('layout').annotate(count=Count('layout'))
        }

        return HttpResponse(params.render_file('am-plugins'))

    @method_decorator(staff_member_required)
    def post(self, request):
        j = json.loads(request.body.decode('utf-8'))

        pluginsMeta = j['pluginsMeta']
        plugins_meta = [
            const.PluginMeta(**kwargs)
            for kwargs in pluginsMeta
        ]

        config.plugins_meta = plugins_meta
        GlobalSetting.set_json('plugins_meta', config.get_plugin_meta_plain(), by=request.user)
        configure_final()

        res = {
            'success': True,
            'comment': '-',
        }

        return HttpResponse(json.dumps(res))



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import traceback

from django.http import HttpResponse

from paipage import config, const

from .params import Params

__all__ = [
    'handler404', 'handler500',
]


def handler404(request, *args, **argv):
    params = Params(request)
    layout = config.template_layout_default + const.HTML_EXT
    params.update({
        'layout_template': layout,
        'title': config.language_available[params['lang']]['error404'],
        'description': '',
    })
    return HttpResponse(params.render_file('ot-output'), status=404)


def handler500(request, *args, **argv):
    params = Params(request)
    layout = config.template_layout_default + const.HTML_EXT
    params.update({
        'layout_template': layout,
        'title': config.language_available[params['lang']]['error500'],
        'description': '',
    })
    return HttpResponse(params.render_file('ot-output'), status=500)



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import json

import jinja2

from django.shortcuts import render
from django.views import View
from django.http import Http404

from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.contrib.admin.views.decorators import staff_member_required
from django.views.decorators.csrf import csrf_exempt

from paipage import config
from paipage.const import PATH_INDEX, HTML_EXT
from paipage.models import Page
from paipage.templating import render_page_full

from .params import Params


class PageView(View):
    def get(self, request, path=PATH_INDEX, lang=None):
        params = Params(request, lang=lang)
        lang = params['lang']

        page = Page.objects.filter(url=path).first()
        if page is None:
            raise Http404

        text = page.get_text(lang)

        if text is None:
            config.logger.error(f'No lang "{lang}" for page "{path}"')
            layout = config.template_layout_default + HTML_EXT
            other_langs = list(page.texts.all())
            params.update({
                'current': page,
                'layout_template': layout,
                'title': config.language_available[params['lang']]['errorNoLang'],
                'description': '',
                'other_langs': other_langs,
            })
            return HttpResponse(params.render_file('ot-nolang'))

        else:
            text.text_full = params.render_str(text.text_full)
            return render_page_full(params, page, text)


@method_decorator(csrf_exempt, name='dispatch')
class PagePreView(View):
    @method_decorator(staff_member_required)
    def post(self, request, ind):
        text = json.loads(request.body.decode('utf-8'))
        params = Params(request)

        page = Page.objects.filter(pk=ind).first()
        if page is None:
            raise Http404

        try:
            text['text_full'] = params.render_str(text['text_full'])
            return render_page_full(params, page, text)
        except jinja2.exceptions.TemplateError as ex:
            params['title'] = 'Template error!'
            params['description'] = repr(ex)
            return HttpResponse(params.render_file('ot-output'))



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import json

from django.views import View
from django.utils.decorators import method_decorator
from django.contrib.admin.views.decorators import staff_member_required
from django.views.decorators.csrf import csrf_exempt
from django.shortcuts import get_object_or_404
from django.http import HttpResponse

from paipage import config
from paipage.const import LANG_NO
from paipage.models import Page, PageText

from .params import Params


class AdminkaStructureView(View):
    @method_decorator(staff_member_required)
    def get(self, request):
        params = Params(request)
        params.scripted['all_pages'] = Page.get_all()
        params.scripted['language_selectable'] = list(config.language_available.keys())
        params.scripted['default_page'] = config.template_page_default
        params.scripted['default_layout'] = config.template_layout_default
        params.scripted['handler_forms'] = {
            key: handler.form_elements
            for key, handler in config.template_handlers.items()
            if handler.form_elements is not None
        }
        params.selectables['template_layout'] = config.template_layout_list
        params.selectables['template_page'] = config.template_page_list
        return HttpResponse(params.render_file('am-struct'))


@method_decorator(csrf_exempt, name='dispatch')
class AdminkaPageView(View):
    @method_decorator(staff_member_required)
    def get(self, request, pk):
        params = Params(request)
        page = get_object_or_404(Page, pk=pk)
        texts = [t.as_dict() for t in page.texts.all()]
        params.scripted['the_page'] = page.as_dict()
        params.scripted['texts'] = {t['language']: t for t in texts}
        params.scripted['language_selectable'] = list(config.language_available.keys())
        params.scripted['default_page'] = config.template_page_default
        params.scripted['default_layout'] = config.template_layout_default
        params.scripted['handler_forms'] = {
            key: handler.form_elements
            for key, handler in config.template_handlers.items()
            if handler.form_elements is not None
        }
        params.selectables['language'] = params.scripted['language_selectable'] + [LANG_NO]
        params.selectables['template_layout'] = config.template_layout_list
        params.selectables['template_page'] = config.template_page_list
        return HttpResponse(params.render_file('am-page'))

    @method_decorator(staff_member_required)
    def post(self, request, pk=-1):
        j = json.loads(request.body.decode('utf-8'))
        if pk == -1:
            page = Page(created_by=request.user)
        else:
            page = get_object_or_404(Page, pk=pk)

        res = {
            'success': True,
            'comment': '-',
        }

        if res['success']:
            new_url = j['url']
            if new_url != page.url:
                other = Page.objects.filter(url=new_url).first()
                if other is not None:
                    res['success'] = False
                    res['comment'] = 'The address is already in use'
                else:
                    page.url = new_url

        if res['success']:
            new_upper = j['upper']
            if new_upper is not None:
                new_upper = int(new_upper)
            print(f'Upper new: {new_upper}, old: {page.upper}')
            if new_upper != page.upper:
                if new_upper == pk:
                    res['success'] = False
                    res['comment'] = 'Cannot link page to itself'
                if new_upper is not None:
                    temp_upper = new_upper
                    while True:
                        if temp_upper is None:
                            break
                        temp = get_object_or_404(Page, pk=temp_upper)
                        if temp.upper:
                            temp_upper = temp.upper.pk
                        else:
                            break
                        if temp_upper == new_upper:
                            break  # prevent looping
                        if pk == temp_upper:
                            res['success'] = False
                            res['comment'] = 'Cannot link page to itself'
                if new_upper is None:
                    page.upper = None
                else:
                    page.upper = get_object_or_404(Page, pk=new_upper)

        if res['success']:
            page.layout = j['template_layout']
            page.layout = '' if page.layout is None else page.layout
            page.template = j['template_page']
            page.template = '' if page.template is None else page.template
            if 'other_settings' in j:
                page.other_settings = json.dumps(j['other_settings'])
            else:
                page.other_settings = '{}'

        if res['success']:
            page.updated_by = request.user
            page.save()
            res['obj'] = page.as_dict()

        return HttpResponse(json.dumps(res))


@method_decorator(csrf_exempt, name='dispatch')
class AdminkaPageTextView(View):
    @method_decorator(staff_member_required)
    def post(self, request, pk, lang):
        # `lang` is old lang, it may be changed
        j = json.loads(request.body.decode('utf-8'))
        page = get_object_or_404(Page, pk=pk)
        text = page.get_text(lang)

        res = {
            'success': True,
            'comment': '-',
        }

        if text is None:
            text = PageText(page=page, created_by=request.user, language=j['language'])
        elif lang != j['language']:
            existing = PageText.objects.filter(page=page, language=j['language']).first()
            if existing is not None:
                res['success'] = False
                res['comment'] = 'Cannot replace existing text language'
            else:
                text.language = j['language']

        if res['success']:
            text.title = j['title']
            text.keywords = j['keywords']
            text.description = j['description']
            text.text_short = j['text_short']
            text.text_full = j['text_full']
            text.updated_by = request.user
            text.save()

        return HttpResponse(json.dumps(res))



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import json

from paipage import config
from paipage.const import LANG_KEY, PATH_INDEX, HTML_EXT
from paipage.models import Page
import paipage.templating as templating


class Params():
    def __init__(self, request, lang=None):
        if lang is None:
            if LANG_KEY in request.session:
                lang = request.session[LANG_KEY]
        if lang not in config.language_available:
            lang = config.language_default
            request.session[LANG_KEY] = lang

        root = Page.objects.filter(url=PATH_INDEX).first()
        if root is None:
            txt = f'No root page found! Create a page for url "{PATH_INDEX}"'
            config.logger.error(txt)
            raise ValueError(txt)

        self.params = {
            'request': request,
            'lang': lang,
            'strings': config.language_available[lang].as_dict(),
            'current': None,
            'sections': list(root.children.all()),
            'layout_template': config.template_layout_default + HTML_EXT,
            'should_show': self._should_show,
            'render_page_short': self._render_page_short,
            'get_handler': self._get_handler,
        }
        self.selectables = {}
        self.scripted = {'selectables': self.selectables}

    def __setitem__(self, key, item):
        self.params[key] = item

    def __getitem__(self, key):
        return self.params[key]

    def _should_show(self, page, text):
        return templating.should_show(self, page, text)

    def _render_page_short(self, page, text):
        return templating.render_page_short(self, page, text)

    def _get_handler(self, page, text):
        return templating.get_handler(self, page, text)

    def update(self, adict):
        self.params.update(adict)

    def prepare(self):
        self.params['scripted'] = {}
        for key in list(self.scripted.keys()):
            self.params['scripted'][key] = json.dumps(self.scripted[key], indent=4, sort_keys=True)
        return self.params

    def render_str(self, template_content):
        template_obj = config.jinja2.from_string(template_content)
        html = template_obj.render(**self.prepare())
        return html

    def render_file(self, template_name):
        template_obj = config.jinja2.get_template(template_name + HTML_EXT)
        html = template_obj.render(**self.prepare())
        return html



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='PaiPage',
    version='1.0.4',
    author='AivanF.',
    author_email='projects@aivanf.com',
    description='A simple CMS with languages and themes',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/AivanF/PaiPage',
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Environment :: Web Environment',
        'Framework :: Django CMS',
        'License :: Freely Distributable',
    ],
)



# Generated by Django 2.2.7 on 2020-05-19 17:46

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('paipage', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='page',
            name='layout',
            field=models.CharField(default='lo-menu-left', max_length=256),
        ),
        migrations.AlterField(
            model_name='page',
            name='template',
            field=models.CharField(default='pg-usual', max_length=256),
        ),
        migrations.AlterField(
            model_name='page',
            name='upper',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='children', to='paipage.Page'),
        ),
        migrations.AlterField(
            model_name='pagetext',
            name='language',
            field=models.CharField(choices=[('_', 'NoLang'), ('en', 'English'), ('ru', 'Русский')], max_length=16),
        ),
        migrations.AlterUniqueTogether(
            name='pagetext',
            unique_together={('page', 'language')},
        ),
    ]



# Generated by Django 2.2.13 on 2020-07-05 05:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('paipage', '0004_globalsetting_updated_by'),
    ]

    operations = [
        migrations.AddField(
            model_name='page',
            name='other_settings',
            field=models.TextField(default='{}'),
        ),
    ]



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import json

import cachetools

from django.db import models
from django.contrib.auth.models import User
from django.http import Http404

from paipage import config
from paipage.const import LANG_NO

__all__ = [
    'Page',
    'PageText',
    'GlobalSetting',
]

dater = '%Y-%m-%d'


class Page(models.Model):
    # Meta info
    upper = models.ForeignKey('self', related_name='children', on_delete=models.PROTECT, null=True, blank=True)

    created = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(User, related_name='created_pages', on_delete=models.PROTECT, null=False)
    updated = models.DateTimeField(auto_now=True)
    updated_by = models.ForeignKey(User, on_delete=models.PROTECT, null=False)

    # Main data
    url = models.CharField(max_length=256, null=False, unique=True)
    public = models.BooleanField(default=True)
    template = models.CharField(default='', max_length=256, null=False, blank=True)
    layout = models.CharField(default='', max_length=256, null=False, blank=True)
    other_settings = models.TextField(default='{}')

    class Meta:
        unique_together = ('upper', 'url',)

    def get_text(self, lang):
        texts = {t.language: t for t in self.texts.all()}
        if lang in texts:
            return texts[lang]
        elif LANG_NO in texts:
            return texts[LANG_NO]
        else:
            return None

    def as_dict(self):
        return {
            'pk': self.pk,
            'url': self.url,
            'public': self.public,
            'template_page': self.template,
            'template_layout': self.layout,
            'upper': self.upper.pk if self.upper else None,
            'other_settings': json.loads(self.other_settings),

            'langs': [],
            'lang_no': False,
            'titles': {},
        }

    @property
    @cachetools.cached(cache=cachetools.Cache(maxsize=32))
    def other_settings_(self):
        return json.loads(self.other_settings)

    @classmethod
    @cachetools.cached(cache=cachetools.TTLCache(maxsize=8, ttl=10))
    def get_all(cls):
        result = {
            p.pk: p.as_dict() for p in Page.objects.all()
        }
        for t in PageText.objects.all():
            if t.language == LANG_NO:
                result[t.page.pk]['lang_no'] = True
            else:
                result[t.page.pk]['langs'].append(t.language)
            result[t.page.pk]['titles'][t.language] = t.title
        return result

    def __str__(self):
        return repr(self)

    def __repr__(self):
        result = f'{self.pk}. Page "{self.url}" {self.updated.strftime(dater)}'\
            f' ({self.texts.count()}/)'\
            f' {"_" if not self.template else self.template }'\
            f' : {"_" if not self.layout else self.layout }'
        return result


language_choices = [(LANG_NO, 'NoLang')] + [
    (x.code, x['name'])
    for x in config.language_available.values()
]


class PageText(models.Model):
    # Meta info
    page = models.ForeignKey(Page, related_name='texts', on_delete=models.PROTECT, null=False)

    created = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(User, related_name='created_texts', on_delete=models.PROTECT, null=False)
    updated = models.DateTimeField(auto_now=True)
    updated_by = models.ForeignKey(User, on_delete=models.PROTECT, null=False)

    # Main data
    language = models.CharField(max_length=16, choices=language_choices)
    title = models.CharField(max_length=256)
    keywords = models.CharField(max_length=256)
    description = models.TextField()
    text_short = models.TextField()
    text_full = models.TextField()

    class Meta:
        unique_together = ('page', 'language',)

    def as_dict(self):
        return {
            'language': self.language,
            'title': self.title,
            'keywords': self.keywords,
            'description': self.description,
            'text_short': self.text_short,
            'text_full': self.text_full,
        }

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'{self.pk}. Text "{self.page.url}" / "{self.language}" {self.updated.strftime(dater)}'


class GlobalSetting(models.Model):
    key = models.CharField(max_length=64, null=False, unique=True)
    value = models.TextField(null=False, blank=True)
    updated = models.DateTimeField(auto_now=True)
    updated_by = models.ForeignKey(User, on_delete=models.PROTECT, null=False)

    @classmethod
    def get(cls, key, default=''):
        obj = cls.objects.filter(key=key).first()
        if obj is not None:
            return obj.value
        else:
            return default

    @classmethod
    def set(cls, key, value, by):
        assert isinstance(value, str)
        assert isinstance(by, User)
        obj = cls.objects.filter(key=key).first()
        if obj is not None:
            obj.value = value
        else:
            obj = cls(key=key, value=value)
        obj.updated_by = by
        obj.save()

    @classmethod
    def get_json(cls, key, default):
        obj = cls.objects.filter(key=key).first()
        if obj is not None:
            return json.loads(obj.value)
        else:
            return default

    @classmethod
    def set_json(cls, key, value, by):
        cls.set(key, json.dumps(value), by=by)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        value = self.value
        if len(value) > 80:
            value = value[:64] + '...'
        return f'{self.pk}. GlobalSetting("{self.key}", {self.updated.strftime(dater)}) = "{value}"'



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

from django.contrib import admin
from . import models

for el in models.__all__:
	admin.site.register(getattr(models, el))



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

default_app_config = 'paipage.apps.PaiPaiConfig'

from .config import config, configure
from .templating import TemplateHandler



__author__ = 'AivanF'
__copyright__ = 'Copyright 2020, AivanF'
__contact__ = 'projects@aivanf.com'

import re

from django.shortcuts import redirect
from django.views import View

from paipage import config
from paipage.const import LANG_KEY


class ChangeLangView(View):
    def get(self, request, lang):
        if lang not in config.language_available:
            lang = config.language_default
        request.session[LANG_KEY] = lang

        next = request.GET.get('next')
        print(f'Next is {next}')
        if next is None:
            next = '/'
        else:
            next = re.sub(r'/[a-z]{2}/', f'/{lang}/', next)
        return redirect(next)
